/* ------------------------------------------------------------------------
System       : cc65 cross-development environment
Platform     : Apple IIgs or Apple IIe with VOC or CB

Bill Buckels
July 2014

This program is an adapted version of my VGA Fan ScreenSaver.

The Algorithm is loosely based on ART.PAS (ART Demonstration Program Version
1.00A 03-18-85) which was shipped with Turbo Pascal Version 3.00 for the IBM
PC. The Author of ART.PAS is Unknown. No credits or Copyright notices appear
in the source code. Maybe it was written by Frank Borland himself.

               .
Licence      : You may use this program for whatever you wish as long
			   as you agree that Bill Buckels has no warranty or
			   liability obligations whatsoever from said use.
------------------------------------------------------------------------ */

#ifndef __APPLE2__
#define __APPLE2__
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <conio.h>

/* the main source file (this file) primarily contains the helper functions */
/* all the SHR core routines etc. are included in the local headers */
#include "shrfan.h"

/* -------------------------------------------------------------- */
/* local random number generator to overcome the lack of a usable */
/* random number generator in cc65...                             */
/* -------------------------------------------------------------- */
/* http://stackoverflow.com/questions/7602919/how-do-i-generate-random-numbers-without-rand-function */
unsigned RandomSeed = (unsigned)0xACE1;
unsigned random()
{
	unsigned bit = ((RandomSeed >> 0) ^ (RandomSeed >> 2) ^ (RandomSeed >> 3) ^ (RandomSeed >> 5) ) & 1;

    RandomSeed =  (RandomSeed >> 1) | (bit << 15);

    return RandomSeed;
}

/* ------------------------------------------------------------- */
/* RandomRange  											     */
/*   returns a random number in a usuable range  (1 to MaxValue) */
/* args                                                          */
/*   MaxValue - the highest number we want                       */
/* ------------------------------------------------------------- */
int RandomRange(int iMaxValue)
{
  int iRetVal;

  do {
    /* get random number */
    iRetVal = random();
    /* get a positive value */
    if (iRetVal < 0) iRetVal *= -1;

  } while(iRetVal < 1);

  /* use modulus of MaxValue if not in range */
  if (iRetVal > iMaxValue)
    iRetVal = (iRetVal%iMaxValue)+1;

  return iRetVal;               /* return a value in range */
}

/* -------------------------------------------------------------- */
/* select new delta values - for geometric fan effect             */
/*   delta values are added during plotting to increment or       */
/*   decrement the end points on the lines to create a fan effect */
/* -------------------------------------------------------------- */
void SelectNewDeltaValues()
{
  iDeltaX1 = RandomRange(DELTAPLUS)-DELTAMINUS;
  iDeltaY1 = RandomRange(DELTAPLUS)-DELTAMINUS;
  iDeltaY2 = RandomRange(DELTAPLUS)-DELTAMINUS;
  iDeltaX2 = RandomRange(DELTAPLUS)-DELTAMINUS;

  /* the increment count controls how far each fan will travel */
  /* before it is replaced with a different geometry           */
  iIncrementCount = (DELTAMULTIPLIER)*(1+RandomRange(DELTASEED));

}

/* ------------------------------------------------------------- */
/* select new color                                              */
/*   called at intervals during plotting to select another color */
/* ------------------------------------------------------------- */
void SelectNewColor()
{

  DrawColor = (unsigned char) RandomRange(HIGHESTCOLOR);

  /* determine how long we will run this color before reselecting */
  iColorCount = (DELTAMULTIPLIER+1)*(1+RandomRange(DELTASEED+1));

}

/* -------------------------------------------------------- */
/* adjust - bounds processing                               */
/*   called to test and reverse x or y term during plotting */
/* -------------------------------------------------------- */
void Adjust(int *iPosition, int *iDelta, int iBoundMax)
{
  int iTest;

  iTest = *iPosition + *iDelta;
  if ((iTest < 1) || (iTest > iBoundMax)) {
    iTest = *iPosition;
    *iDelta = -*iDelta;
  }
  *iPosition = iTest;
}


/* ----------------- */
/* wander the lines  */
/* ----------------- */
void WanderingLines(void)
{

  WORD wCnt;

  for (wCnt = 0; wCnt < wNumberOfLines; wCnt++) {
 	/* erase an old line (if any) before losing values     */
  	/* the line array is a ring buffer using FIFO to erase */
   	/*   the oldest line before drawing a new one          */
  	if (wXBound == XBOUND)
    	line320(WanderLine[iCurrentLine].iLX1, WanderLine[iCurrentLine].iLY1,
            	WanderLine[iCurrentLine].iLX2, WanderLine[iCurrentLine].iLY2,0);
  	else
    	doublewide320(WanderLine[iCurrentLine].iLX1, WanderLine[iCurrentLine].iLY1,
            	WanderLine[iCurrentLine].iLX2, WanderLine[iCurrentLine].iLY2,0);

    /* end of color term so select a new color */
    if (iColorCount < 1) SelectNewColor();

    /* end of shape so get new values for our geometry */
    if (iIncrementCount < 1) SelectNewDeltaValues();

    /* get values for the current line */
    /* if we have collided with the sides */
    /*   then change direction */

    Adjust(&iLX1, &iDeltaX1, wXBound);
    Adjust(&iLY1, &iDeltaY1, YBOUND);
    Adjust(&iLX2, &iDeltaX2, wXBound);
    Adjust(&iLY2, &iDeltaY2, YBOUND);

    /* -------------------------------------------------- */
    /* save the current line into the line structure      */
    /*   restored in black when required again            */
    /* -------------------------------------------------- */
    WanderLine[iCurrentLine].iLX1 = iLX1;
    WanderLine[iCurrentLine].iLY1 = iLY1;
    WanderLine[iCurrentLine].iLX2 = iLX2;
    WanderLine[iCurrentLine].iLY2 = iLY2;
    WanderLine[iCurrentLine].color = DrawColor;

    if (wXBound == XBOUND)
    	line320(WanderLine[iCurrentLine].iLX1, WanderLine[iCurrentLine].iLY1,
            	WanderLine[iCurrentLine].iLX2, WanderLine[iCurrentLine].iLY2,
            	WanderLine[iCurrentLine].color);
    else
    	doublewide320(WanderLine[iCurrentLine].iLX1, WanderLine[iCurrentLine].iLY1,
            	WanderLine[iCurrentLine].iLX2, WanderLine[iCurrentLine].iLY2,
            	WanderLine[iCurrentLine].color);

    /* get ready for the next pass... and increment and decrement
         animation counters as required */
    iCurrentLine++;
    if (iCurrentLine > (int)wSerpentLength)
      iCurrentLine = 0;
    iColorCount--;
    iIncrementCount--;
  }
  return;
}

/* ------------------------------------------- */
/* regenerate the demo when RETURN is pressed. */
/* ------------------------------------------- */
void ReGenerate()
{
	int idx;

	/* start somewhere new each time */
	iLX1 = iLY1 = iLX2 = iLY2 = RandomRange(YBOUND);

  	/* reset other values */
  	iColorCount = iIncrementCount = iCurrentLine = NIL;

 	/* zero-out previous line coordinates */
	for (idx = 0; idx < MEMORYMAX; idx++) {
    	WanderLine[idx].color = 0;

    	WanderLine[idx].iLX1 =
    	WanderLine[idx].iLY1 =
    	WanderLine[idx].iLX2 =
    	WanderLine[idx].iLY2 = NIL;
  	}

	random(); /* shake-it-up */

    /* clear screen and restore initial palette */
    clearpalette();
    clear320(LOBLACK);
    memcpy(&paletteptr[0],&rgbsaved[0],48);
    setpalette(paletteptr,0);


    /* good to go now! */
}

/* ----------------------------------------------------------- */
/* slide colors 1-15 sideways when the ARROW keys are pressed. */
/* ----------------------------------------------------------- */
void ShufflePalette(char direction)
{
	random(); /* shake-it-up */

	switch(direction) {
		case RTARROW:
		case DNARROW:
		case '+':
    		/* forward - leave black alone */
			memcpy(&rgbwork[0],&paletteptr[3],45);
			memcpy(&paletteptr[6],&rgbwork[0],42);
			memcpy(&paletteptr[3],&rgbwork[42],3);
			break;
		case LTARROW:
		case UPARROW:
	    case '-':
    		/* reverse - leave black alone */
			memcpy(&rgbwork[0],&paletteptr[3],45);
			memcpy(&paletteptr[3],&rgbwork[3],42);
			memcpy(&paletteptr[45],&rgbwork[0],3);
			break;
		default:
		    return;
	}
    setpalette(paletteptr,0);
    /* good to go now! */
}


/* ------------------------------------------------------ */
/* save the SHR screen to disk when the S key is pressed. */
/* ------------------------------------------------------ */
/* a simple $C1 $0000 SHR image save routine */
int picsave(char *name)
{

	FILE *fp;
	int i, c, status = ERR_WRITEPIXELS;
	unsigned src = 0x2000;	/* set src to scanline 0 */
    extern unsigned char _filetype;
    extern unsigned _auxtype;

    /* remove file if it already exists */
    fp = fopen(name,"rb");
    if (fp != NULL) {
		fclose(fp);
		remove(name);
	}

    /* set file type to Screen Image */
    _filetype = 0xC1;
    _auxtype = 0x0000;

	fp = fopen(name,"wb");
    if (fp == NULL) return ERR_OPEN;

	for (;;) {

		/* write image data - 32000 bytes */
		for (i=0;i<4;i++) {
		   /* move to main memory */
		   auxtomain(src,src+7999,0x2000);
		   c = fwrite((char *)0x2000,1,8000,fp);
		   if (c != 8000)break;
		   src += 8000;
		}
		if (c!=8000) break;

        /* write scbs and palette - 768 bytes */
		status = ERR_WRITEPALETTE;
		auxtomain(src,src+767,0x2000);
		c = fwrite((char *)0x2000,1,768,fp);
		if (c!=768) break;
		status = SUCCESS;
		break;
	}

    /* close file and remove if a write error occurred */
    /* likely out of disk space */
	fclose(fp);
	if (status != SUCCESS) remove(name);
	return status;
}

/* check for existing saved screen snapshot files
   we are supporting 100 using automatic naming
   this is a first-fit so if there are gaps in the
   existing file names the gaps will be filled */
int GetNextPicName(char *name)
{
	FILE *fp;
	int i;

	/* SHRFAN00.SHR to SHRFAN99.SHR */
	for (i = 0; i < 100; i++) {
		sprintf(name,"SHRFAN%02d.SHR",i);
	    fp = fopen(name,"rb");
	    if (fp == NULL) return SUCCESS;
	    fclose(fp);
	}
	/* if we already have 100 files that's enough */
	name[0] = (char)0;
	return ERR_EXCESSIVE;
}

int main(void)
{
	unsigned char ch, buf[40];
    int status = SUCCESS;

    /* initialize text mode. stay in 40 column mode. */
    texton();
    clrscr();

    /* initialize empty palette and scb's */
    initbuffers();

    /* clear keyboard buffer */
    while (kbhit() != 0) {
		cgetc();
		random();
	}

    /* settings screen */
    puts("shrfan(C) Bill Buckels 2014.");
    puts("--------------------------------------");
    puts("Settings - press RETURN for defaults:");
    puts("Serpent Length Default: 150 Lines");
    puts("Lines Between KeyCheck: 2 Lines");
    puts("Pixel Size Default    : Single Pixels");
    puts("--------------------------------------");
    puts("Enter Serpent Length (50-200):");
    gets(buf);
    wSerpentLength = (WORD) atoi(buf);
    if (wSerpentLength < MEMORYMIN)
    	wSerpentLength  = MEMORYMEDIUM;
  	else if(wSerpentLength > MEMORYSIZE)
    	wSerpentLength  = MEMORYMEDIUM;
    puts("Enter Keycheck (1 or more Lines):");
    gets(buf);
    wNumberOfLines = (WORD) atoi(buf);
  	if (wNumberOfLines <  MINLINES)
    	wNumberOfLines = DEFLINES;
  	else if (wNumberOfLines >  wSerpentLength)
    	wNumberOfLines = wSerpentLength;
    puts("Do you want Double Pixels (Y/N)?");
    while (kbhit() == 0) random();
    ch = toupper(cgetc());
    if (ch == 'Y') wXBound = XPAIR;

	/* clear keyboard buffer */
	while (kbhit() != 0) {
		cgetc();
		random();
	}

    /* KeyPress Info and Palette Selection Screen */
    clrscr();
    puts("shrfan(C) Bill Buckels 2014.");
    puts("--------------------------------------");
    puts("Demo Keys:");
    puts("SPACEBAR   - Pause-Resume Toggle");
    puts("  S - Save Paused Screen SHRFANxx.SHR");
    puts("ARROW Keys - Move Colors Left or Right");
    puts("RETURN Key - Refresh and Reset");
    puts("ESC Key    - Exit");
    puts("--------------------------------------");
    puts("Select a 16-color palette:");
    puts("1 - Kegs32 (default)");
    puts("2 - CiderPress");
    puts("3 - Old AppleWin");
    puts("4 - New AppleWin");
    while (kbhit() == 0) random();
    ch = cgetc();

	/* turn shr on */
	shgron();
	/* now that the shr display is in auxiliary memory */
	/* clear the palette and point all the scbs to the
	   first palette */
    clearpalette();
    clearscbs();
	/* clear the shr screen to black */
    clear320(LOBLACK);
    /* set the palette that the user has selected */
    switch(ch) {
		case '2' : paletteptr = (unsigned char *)&rgbciderpress[0];break;
		case '3' : paletteptr = (unsigned char *)&rgbawinold[0];break;
		case '4' : paletteptr = (unsigned char *)&rgbawinnew[0];break;
		default  : paletteptr = (unsigned char *)&rgbkegs32[0];break;
	}
    setpalette(paletteptr,0);
	/* save a copy of the user's palette selection for resetting if needed */
	memcpy(&rgbsaved[0],&paletteptr[0],48);

    /* clear keyboard buffer */
    while (kbhit() != 0) {
		cgetc();
		random();
	}
	for (;;) {
		WanderingLines();
		if (kbhit() != 0) {
			ch = toupper(cgetc());
			/* color change if arrow key is pressed */
			ShufflePalette(ch);
			while (kbhit() != 0) {
				cgetc();
				random();
			}
			if (ch == ESCAPE) break;
			else if (ch == RETURNKEY)ReGenerate();
			else if (ch == SPACEBAR || ch == 'S') {
				/* if they pressed the S key before pausing they will
				   need to press it again to save */
				for (;;) {
					while (kbhit() == 0) random();
					ch = toupper(cgetc());
					/* the spacebar and the return key have the same effect
					   when paused both resume but pressing the return key
					   clears the screen and regenerates the demo */
					if (ch == SPACEBAR) break;
					else if (ch == RETURNKEY) {
						ReGenerate();
						break;
					}
					else if (ch == 'S') {
						/* not reporting errors here */
						/* if we had an error saving before just skip this */
						if (status == SUCCESS) status = GetNextPicName((char *)&buf[0]);
						if (status == SUCCESS) status = picsave(buf);
					}
					while (kbhit() != 0) {
						cgetc();
						random();
					}
					if (ch == 'S') break;
					/* when paused also allow color changes */
					ShufflePalette(ch);
				}
			}
		}
	}

	/* clear the palette and scbs then
	   turn shroff and re-initialize text mode */
	clearpalette();
	clearscbs();
	shgroff();
	texton();
	clrscr();

	/* report the error if a problem saving the screenshot */
	if (status != SUCCESS) {
		switch(status) {
		    case ERR_OPEN:         printf("Error Opening %s\n",buf); break;
		    case ERR_WRITEPIXELS:
		    case ERR_WRITEPALETTE: printf("Error Writing %s\n",buf); break;
		    case ERR_EXCESSIVE:    puts("Error: 100 ScreenShot Limit Exceeded!"); break;
		    default: printf("Unknown Error: %d!\n",status);
		}
		cgetc();
	}

    /* and exit */
    return SUCCESS;

}