/* ------------------------------------------------------------------------
System       : cc65 cross-development environment
Platform     : Apple IIgs or Apple IIe with VOC or CB

Bill Buckels
July 2014

Licence      : You may use this program for whatever you wish as long
			   as you agree that Bill Buckels has no warranty or
			   liability obligations whatsoever from said use.
------------------------------------------------------------------------ */
#include "shr.h"

/* some local constants for status and return codes */
#define SUCCESS 0
#define ERR_OPEN         -1
#define ERR_WRITEPIXELS  -2
#define ERR_WRITEPALETTE -3
#define ERR_EXCESSIVE    -4

/* some local constants for kepress values */
#define RTARROW   (unsigned char)21
#define LTARROW   (unsigned char)8
#define UPARROW   (unsigned char)11
#define DNARROW   (unsigned char)10
#define ESCAPE    (unsigned char)27
#define DELETE    (unsigned char)127
#define RETURNKEY (unsigned char)13
#define SPACEBAR  (unsigned char)32

/* local constants for the lo-res colors and color order */
/* these are also the double lo-res and double hi-res colors */
#define LOBLACK   	0
#define LORED     	1
#define LODKBLUE 	2
#define LOPURPLE  	3
#define LODKGREEN	4
#define LOGRAY    	5
#define LOMEDBLUE	6
#define LOLTBLUE 	7
#define LOBROWN   	8
#define LOORANGE  	9
#define LOGREY    	10
#define LOPINK    	11
#define LOLTGREEN	12
#define LOYELLOW  	13
#define LOAQUA    	14
#define LOWHITE   	15
#define MAXCOLORS   16

/* SHR Fan Settings etc. */
#define NIL             0
#define MEMORYSIZE      200
#define MEMORYMIN       50
#define MEMORYMEDIUM    150
#define MEMORYMAX       (MEMORYSIZE + 1)
#define DEFLINES        2
#define MINLINES        1

#define HIGHESTCOLOR    15
#define DELTAPLUS       16
#define DELTAMINUS      8
#define DELTAMULTIPLIER 8
#define DELTASEED       18

/* SHR screen boundaries */
#define XBOUND 319
#define XPAIR  159
#define YBOUND 199

/* Windows 3.1 style typedef to preserve referential integrity with
   code ported from windows - makes visual comparison easier (for me at least) */
typedef unsigned int  WORD;

WORD wNumberOfLines    = DEFLINES;        /* minimum number of lines */
WORD wSerpentLength    = MEMORYMEDIUM;    /* length parameter */
WORD wXBound           = XBOUND;          /* pixel or pixel-pair */

/* global SHR color index for palette 0 colors 1-15 */
unsigned char DrawColor = (unsigned char)LOBLACK;

/* global coordinates used to compute values for animation */
int iLX1 = NIL, iLX2 = NIL, iLY1 = NIL, iLY2 = NIL,
           iCurrentLine = NIL, iColorCount = NIL, iIncrementCount = NIL,
           iDeltaX1 = NIL, iDeltaY1 = NIL, iDeltaX2 = NIL, iDeltaY2 = NIL;


/* a FIFO ring buffer to store line history */
struct {
  int iLX1, iLY1;
  int iLX2, iLY2;
  unsigned char color;
  } WanderLine[MEMORYMAX];


/* 4 - palette options */
/* these are the rgb values of the lo-res colors from
   the AppleWin and Kegs32 Emulators and from
   the CiderPress File Viewer.

   two sets of AppleWin Colors are provided:
   one from an older version and one from a newer version.

   the routines in this demo are based on a fixed palette
   so it was convenient to use known colors and color order.

   these palettes should really be two dimensional arrays
   but the extra curly braces required by ISO C make these
   less readable so instead of rgb[16][3] I am using rgb[48].

   Ansi C isn't so fussy and neither is K&R C.

*/

unsigned char rgbsaved[48], rgbwork[48], *paletteptr;

unsigned char rgbawinold[48] = {
	0,0,0,       /* black    */
	208,0,48,	 /* red      */
	0,0,128,	 /* dk blue  */
	255,0,255,	 /* purple   */
	0,128,0,	 /* dk green */
	128,128,128, /* gray     */
	0,0,255,	 /* med blue */
	96,160,255,	 /* lt blue  */
	128,80,0,	 /* brown    */
    255,128,0,	 /* orange   */
    192,192,192, /* grey     */
    255,144,128, /* pink     */
    0,255,0,	 /* lt green */
    255,255,0,	 /* yellow   */
    64,255,144,	 /* aqua     */
    255,255,255};/* white    */

unsigned char rgbawinnew[48] = {
	0,0,0,       /* black    */
	157,9,102,   /* red      */
	42,42,229,   /* dk blue  */
	199,52,255,  /* purple   */
	0,118,26,    /* dk green */
	128,128,128, /* gray     */
	13,161,255,  /* med blue */
	170,170,255, /* lt blue  */
	85,85,0,     /* brown    */
    242,94,0,    /* orange   */
    192,192,192, /* grey     */
    255,137,229, /* pink     */
    56,203,0,    /* lt green */
    213,213,26,  /* yellow   */
    98,246,153,  /* aqua     */
    255,255,254};/* white    */

unsigned char rgbciderpress[48] = {
	0,0,0,		 /* black    */
	221,0,51,	 /* red      */
	0,0,153,	 /* dk blue  */
	221,34,221,	 /* purple   */
	0,119,34,	 /* dk green */
	85,85,85,	 /* gray     */
	34,34,255,	 /* med blue */
	102,170,255, /* lt blue  */
	136,85,0,	 /* brown    */
	255,102,0,	 /* orange   */
	170,170,170, /* grey     */
	255,153,136, /* pink     */
	17,221,0,	 /* lt green */
	255,255,0,	 /* yellow   */
	68,255,153,	 /* aqua     */
	255,255,255};/* white    */

unsigned char rgbkegs32[48] = {
	0,0,0,		 /* black    */
	221,0,51,	 /* red      */
	0,0,153,	 /* dk blue  */
	221,0,221,	 /* purple   */
	0,119,0,	 /* dk green */
	85,85,85,	 /* gray     */
	34,34,255,	 /* med blue */
	102,170,255, /* lt blue  */
	136,85,34,	 /* brown    */
	255,102,0,	 /* orange   */
	170,170,170, /* grey     */
	255,153,136, /* pink     */
	0,221,0,	 /* lt green */
	255,255,0,	 /* yellow   */
	0,255,153,	 /* aqua     */
	255,255,255};/* white    */

unsigned char scanline[960],crt[1],shrpal[32];

#pragma optimize (push,off)
void putpixel320(unsigned x,unsigned y, unsigned char color)
{
    unsigned *src  = (unsigned *)0x3c;
    unsigned *dest = (unsigned *)0x42;
	unsigned pixel, buf = (unsigned) &crt[0];

    /* return if out of range */
    if (x > 319 || y > 199) return;

    pixel = (unsigned) (0x2000 + (y * 160) + (x/2));

	/* move pixel to buffer */
	src[0] = src[1] = pixel;
	dest[0] = buf;
	asm("clc");
	asm("jsr $c311");

	if (x%2 == 1) {
		/* odd pixels in low nibble */
		/* mask value to preserve high nibble */
		crt[0] &= (unsigned char) 240;
		crt[0] |= (unsigned char) color;
	}
	else {
		/* even pixels in high nibble */
		/* mask value to preserve low nibble */
		crt[0] &= (unsigned char) 15;
		crt[0] |= (unsigned char)(color << 4);
	}

	/* move buffer to pixel */
	src[0] = src[1] = buf;
	dest[0] = pixel;
	asm("sec");
	asm("jsr $c311");
}
#pragma optimize (pop)


/* plots pixel pairs using single-buffering */
#pragma optimize (push,off)
void putpairs320(unsigned x, unsigned y, unsigned pairs, unsigned char color)
{
    unsigned *src  = (unsigned *)0x3c;
    unsigned *dest = (unsigned *)0x42;
	unsigned buf = (unsigned) &scanline[0];

    if (y>199) return;

    /* set pair color */
    if (color < 16)memset((char *)&scanline[0],(char)(color << 4 | color),pairs);

	/* move buffer to shr display */
	src[0] = buf;
	src[1] = buf + pairs - 1;
	dest[0] = (unsigned) (0x2000 + (y * 160) + (x/2));
	asm("sec");
	asm("jsr $c311");
}
#pragma optimize (pop)

void hline320(unsigned y,unsigned x1, unsigned x2, unsigned char color)
{
	unsigned temp, pairs;

    /* swap co-ordinates if out of order */
    if (x1>x2) {
		temp = x1;
		x1   = x2;
		x2   = temp;
	}

	if (y > 199 || x1 > 319) return;
	if (x2 > 319) x2=319;

	if (x1 == x2) {
		putpixel320(x1,y,color);
		return;
	}

	if ((x1%2) == 1) {
		putpixel320(x1,y,color);
		x1++;
	}
	if ((x2%2) == 0) {
		putpixel320(x2,y,color);
		x2--;
	}

	pairs = ((x2 + 1) - x1) / 2;

	if (pairs > 0) {
    	putpairs320(x1,y,pairs,color);
	}
}

/* simple vertical line - must set individual pixels */
void vline320(unsigned x,unsigned y1, unsigned y2, unsigned char color)
{
	unsigned temp;

    /* swap co-ordinates if out of order */
    if (y1>y2) {
		temp = y1;
		y1   = y2;
		y2   = temp;
	}

    y2++;
    while (y1 < y2) {
	    putpixel320(x,y1,color);
		y1++;
	}

}


/* Bresenham Algorithm line drawing routine for SHR mode320 */
void line320(int x1, int y1, int x2, int y2, unsigned char color)
{

	int dx, dy, sx, sy, err, err2;

    /* single pixel */
    if (x1 == x2 && y1 == y2) {
		putpixel320((unsigned)x1,(unsigned)y1,color);
		return;
	}

	/* vertical line */
	if (x1 == x2) {
		vline320((unsigned)x1, (unsigned)y1, (unsigned)y2, color);
		return;

	}

	/* horizontal line */
	if (y1 == y2) {
		hline320((unsigned) y1, (unsigned)x1, (unsigned)x2, color);
		return;
	}

    /* vector */
	if(x1 < x2) {
		dx = x2 - x1;
		sx = 1;
	}
	else {
		sx = -1;
		dx = x1 - x2;
	}

	if(y1 < y2) {
		sy = 1;
		dy = y2 - y1;
	}
	else {
		sy = -1;
		dy = y1 - y2;
	}

	err = dx-dy;

	for (;;) {
		putpixel320((unsigned)x1,(unsigned)y1,color);

		if(x1 == x2 && y1 == y2)break;

		err2 = err*2;

		if(err2 > (0-dy)) {
			 err = err - dy;
			 x1 = x1 + sx;
		}

		if(err2 <  dx) {
			 err = err + dx;
			 y1 = y1 + sy;
		}
	}
}


/* Bresenham Algorithm doublewide line drawing routine for SHR mode320 */
/* resolution is 160 x 200 rather than 320 x 200 */
void doublewide320(int x1, int y1, int x2, int y2, unsigned char color)
{

	int dx, dy, sx, sy, err, err2;

    /* single pixel */
    if (x1 == x2 && y1 == y2) {
		putpairs320((unsigned)x1*2,(unsigned)y1,1,color);
		return;
	}

	/* vertical line */
	if (x1 == x2) {
		/* swap y co-ordinates if out of order */
		if (y1 > y2) {
			sy = y1;
			y1 = y2;
			y2 = sy;
		}
		x1 *= 2;
		y2++;
		while (y1 < y2) {
		  	putpairs320((unsigned)x1,(unsigned)y1,1,color);
		  	y1++;
	  	}
		return;
	}

	/* horizontal line */
	if (y1 == y2) {
		if (x1 > x2) {
			sx = x1;
			x1 = x2;
			x2 = sx;
		}
	    sx = ((x2 + 1) - x1);
	    putpairs320((unsigned)x1*2,(unsigned)y1,sx,color);
		return;
	}

    /* vector */
	if(x1 < x2) {
		dx = x2 - x1;
		sx = 1;
	}
	else {
		sx = -1;
		dx = x1 - x2;
	}

	if(y1 < y2) {
		sy = 1;
		dy = y2 - y1;
	}
	else {
		sy = -1;
		dy = y1 - y2;
	}

	err = dx-dy;

	for (;;) {
        putpairs320((unsigned)x1*2,(unsigned)y1,1,color);

		if(x1 == x2 && y1 == y2)break;

		err2 = err*2;

		if(err2 > (0-dy)) {
			 err = err - dy;
			 x1 = x1 + sx;
		}

		if(err2 <  dx) {
			 err = err + dx;
			 y1 = y1 + sy;
		}
	}
}

/* clear the SHR screen to color
   probably no good if unknown palettes are used
   mostly for single-palette pixel graphics and the like */
void clear320(unsigned char color)
{
	unsigned char paircolor = (unsigned char) (color << 4 | color);
	unsigned y;
	unsigned src1 = (unsigned) &scanline[0], src2 = src1+159;

	/* put pixel pair array on SHR display */
	/* fill scanline buffer with color pairs */
    memset((char *)&scanline[0],paircolor,160);
    /* wipe-down 200 scan lines with color */
    for (y=0;y<200;y++) {
		maintoaux(src1,src2,(unsigned) (0x2000 + (y * 160)));

	}
}

/* sets a 12 bit color palette line from an array of 24 bit color values */
/* rgbpal is a 48 byte character array of 16 - r,g,b values */
/* palidx is the SHR palette number in the range of 0-15 */
void setpalette(unsigned char *rgbpal,int palidx)
{
	unsigned char r,g,b;
	int i,j,k;
	unsigned src1 = (unsigned)&shrpal[0], src2 = src1 + 31;

	/* build 12 bit palette line of $0RGB color entries
	   from 24 bit r,g,b values */
	for (i=0,j=0,k=0;i<16;i++,j+=3,k+=2) {
		r = rgbpal[j]   >> 4;
		g = rgbpal[j+1] >> 4;
		b = rgbpal[j+2] >> 4;
		shrpal[k] = (unsigned char)((g << 4) | b);
		shrpal[k+1] = r;
	}

    /* move palette line to palette */
    maintoaux(src1, src2,(unsigned)(0x9e00 + (palidx * 32)));

}
